#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
#
# File:            deCONZRESTAPI.py
#
# License:         Donationware, see attached LICENSE file for more
#                  information
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         1.8b
#
# Date:            23.01.2024
#
# Description:     Library to communicate with the deCONZ REST-API.
#
# Supported devices:
#                  Zigbee devices through deCONZ Zigbee controller.
#
# Requirements:    Python interpreter 2.7 or newer (www.python.org)
#
# Version history: ** 23.01.2024 v1.8b (Olli Lammi) **
#                  First version.
#
###########################################################################

# Imports

import string
import json
import urllib2

from modules.core import log


###########################################################################

# Constants


TYPE_CONFIG = 'config'
TYPE_SENSORS = 'sensors'

SENSOR_TYPES = { \
    'ZHATemperature': {'temperature': {'factor': 0.01, 'format': '%.2f', 'unit': 'deg'}}, \
    'ZHAHumidity': {'humidity': {'factor': 0.01, 'format': '%.2f', 'unit': '%'}}, \
    'ZHAPressure': {'pressure': {'factor': 1, 'format': '%d', 'unit': 'hPa'}}, \
    'ZHAPower': {'current': {'factor': 1, 'format': '%d', 'unit': 'mA'}, \
                 'power': {'factor': 1, 'format': '%d', 'unit': 'W'}, \
                 'voltage': {'factor': 1, 'format': '%d', 'unit': 'V'}}, \
    'ZHAConsumption': {'consumption': {'factor': 1, 'format': '%d', 'unit': 'Wh'}, \
                       'power': {'factor': 1, 'format': '%d', 'unit': 'W'}}, \
    'ZHALightLevel': {'lux': {'factor': 1, 'format': '%d', 'unit': 'lux'}, \
                      'lightlevel': {'factor': 1, 'format': '%d', 'unit': 'lux'}}, \
    'ZHAThermostat': {'temperature': {'factor': 0.01, 'format': '%.2f', 'unit': 'deg'}} \
}



###########################################################################

# Classes

class DeCONZRESTAPI(log.Logging):
    def __init__(self):
        log.Logging.__init__(self, 'DeCONZRESTAPI')
        self.url = ''
        self.apikey = ''

    def runQueryCommand(self, cmds):
        res = {}
        data_sensors = self.fetchData(TYPE_SENSORS)

        if not data_sensors:
            self.Log("ERROR: No data.")
            return {}

        for cmd in cmds:
            temp = string.split(cmd, '/')
            if len(temp) != 4:
                self.Log("ERROR: DECONZ MEASURE point has to have exactly 4 fields separated by /. Skipping...")
                continue
            if temp[0].upper().strip() == TYPE_SENSORS.upper():
                tempname = temp[1].strip()
                temptype = temp[2].strip()
                tempfield = temp[3].strip()
                if temptype not in SENSOR_TYPES:
                    self.Log("ERROR: Unknown sensor type in MEASURE command: %s" % temptype)
                    continue
                if tempfield not in SENSOR_TYPES[temptype]:
                    self.Log("ERROR: Unknown sensor state field type in MEASURE command: %s/%s" % (temptype, tempfield))
                    continue
                tempattrib = SENSOR_TYPES[temptype][tempfield]

                if data_sensors:
                    for id in data_sensors:
                        tempsensor = data_sensors[id]
                        if 'name' in tempsensor and 'type' in tempsensor and 'state' in tempsensor:
                            if tempsensor['name'] == tempname and tempsensor['type'] == temptype and tempfield in tempsensor['state']:
                                res[cmd] = tempattrib['format'] % (tempsensor['state'][tempfield] * tempattrib['factor'])
            else:
                self.Log("ERROR: Unknown data type in MEASURE command: %s" % temp[0])

        return res

    def initAPI(self):
        data = self.fetchData(TYPE_CONFIG)
        if data:
            devname = ''
            fwversion = ''
            modelid = ''
            name = ''
            mac = ''
            if 'devicename' in data:
                devname = data['devicename']
            if 'fwversion' in data:
                fwversion = data['fwversion']
            if 'modelid' in data:
                modelid = data['modelid']
            if 'name' in data:
                name= data['name']
            if 'mac' in data:
                mac = data['mac']

            self.Log("Detected deCONZ REST-API: %s %s %s %s (%s)" % (devname, fwversion, modelid, mac, name))

            data = self.fetchData(TYPE_SENSORS)
            if data:
                self.Log("Detected sensors:")
                for id in data:
                    tempsensor = data[id]
                    tempname = ''
                    temptype = ''
                    if 'name' in tempsensor:
                        tempname = tempsensor['name']
                    if 'type' in tempsensor:
                        temptype = tempsensor['type']
                    battery = ''
                    tempstat = ''
                    if 'config' in tempsensor and 'battery' in tempsensor['config']:
                        battery = str(tempsensor['config']['battery'])
                    if battery:
                        tempstat = tempstat + ' Battery: %s' % battery
                    self.Log("    sensor/%s/%s -%s" % (tempname, temptype, tempstat))
            return True
        return False

    def fetchData(self, type = TYPE_SENSORS):
        tempurl = self.url
        if tempurl[len(tempurl)-1] != '/':
            tempurl = tempurl + '/'
        tempurl = tempurl + self.apikey + '/' + type
        self.Debug("Getting data using URL: %s" % tempurl)

        data = None
        try:
            f = urllib2.urlopen(tempurl)
            data = f.read()
            f.close()
        except:
            data = None

        if not data or len(data) <= 0:
            self.Debug("ERROR: Cannot get data.")
            return None

        try:
            data = json.loads(data)
            self.Debug("Retreived data: %s" % (repr(data)))
        except:
            self.Debug("ERROR: Error decoding JSON data from the received data.")
            return None

        return data
