#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            sourceEnerventEda.py
#
# License:         Donationware, see attached LICENSE file for more 
#                  information
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         1.7j
#
# Date:            01.02.2017
#
# Description:     TaloLogger configuration wrapper for Enervent EDA source.
#                  
# Version history: ** 14.01.2015 v1.6j (Olli Lammi) **
#                  Added to taloLogger. 
#
#                  ** 31.01.2017 v1.7j (Olli Lammi) **
#                  Added Modbus master mode to serial module. 
#
#                  ** 01.02.2017 v1.7j (Olli Lammi) **
#                  Added Modbus off-by-one fix. 
#
###########################################################################

# Imports

import string

from modules.core import log
from modules.core import configuration
from modules.core import dataSource

from modules.datasources.modbus import sourceModbus 

import enerventEda

###########################################################################

# Constants

EDA_TYPE_MAP = { \
                 enerventEda.TYPE_INT16: ['INT16', '1'], \
                 enerventEda.TYPE_UINT8: ['UINT8', '1'], \
                 enerventEda.TYPE_UINT16: ['UINT16', '1'], \
                 enerventEda.TYPE_INT16_10: ['FINT16', '0.1'], \
                 enerventEda.TYPE_INT16_1000: ['FINT16', '0.001'] }

###########################################################################

# Classes

class EdaModbusTCPConf(log.Logging, configuration.Configurable, dataSource.DataSource):
    def __init__(self, modname):
        log.Logging.__init__(self, 'EdaModbusTCPConf')
        configuration.Configurable.__init__(self)
        self.setModuleName(modname)
        dataSource.DataSource.__init__(self)
        self.setID(modname)
        
        self.privateConf = None
        self.modbusSource = None
        self.hostName = None
        self.tcpPort = None
        self.modbusUnitId = 1
        self.offbyonefix = False
        
    def getServiceName(self):
        return self.getModuleName()

    def handleConfiguration(self, conf):
        self.hostName = conf.getValue('HOSTNAME', '', self.getModuleName())
        self.tcpPort = conf.getValue('TCPPORT', '502', self.getModuleName())
        try:
            self.modbusUnitId = int(conf.getValue('UNITID', '1', self.getModuleName()))
        except:
            return (-1, 'Cannot convert UNITID to integer.')

        self.offbyonefix = conf.isTrue('OFFBYONEFIX', self.getModuleName())

        self.privateConf = configuration.Configuration()
        self.privateConf.setValue('TYPE', 'TCP', self.getModuleName())
        self.privateConf.setValue('HOSTNAME', self.hostName, self.getModuleName())
        self.privateConf.setValue('TCPPORT', self.tcpPort, self.getModuleName())
        self.privateConf.setValue('RESPONSETIMEOUT', '3.0', self.getModuleName())

        coils = []        
        holdingregisters = []        
        for register in enerventEda.EDA_DEVICES['DEFAULT']:
            if self.offbyonefix:
                address = register[1] + 1
            else:
                address = register[1]
            edaType = EDA_TYPE_MAP[register[3]]
            if register[2] == enerventEda.TYPE_COIL:
                coils.append("%s:%d:%d" % (register[0], self.modbusUnitId, address))
            elif register[2] == enerventEda.TYPE_HREG:
                holdingregisters.append("%s:%d:%d:%s:%s" % (register[0], self.modbusUnitId, address, edaType[0], edaType[1]))
        
        self.privateConf.setValue('COILSTATE', coils, self.getModuleName())
        self.privateConf.setValue('HOLDINGREGISTER', holdingregisters, self.getModuleName())

        return (1, '')
            
    def initConfiguration(self):        
        self.modbusSource = sourceModbus.ModbusConf(self.getModuleName())

        (stat, msg) = self.modbusSource.handleConfiguration(self.privateConf)
        if stat != 1:
            return (stat, msg)
        
        (stat, msg) = self.modbusSource.initConfiguration()
        if stat != 1:
            return (stat, msg)
                
        return (1, '')

    def releaseConfiguration(self):
        if self.modbusSource != None:
            self.modbusSource.releaseConfiguration()
            self.modbusSource = None
        return (1, '')

    def runDataSourceQueryCommandImpl(self, cmds):
        return self.modbusSource.runDataSourceQueryCommandImpl(cmds)

    @staticmethod
    def getModuleTypeName():
        return 'EDAMODBUSTCP'

    @staticmethod
    def getAllowedConfigurationKeys():
        return (['HOSTNAME', 'TCPPORT', 'UNITID', 'OFFBYONEFIX'], [])


class EdaModbusSerialConf(log.Logging, configuration.Configurable, dataSource.DataSource):
    def __init__(self, modname):
        log.Logging.__init__(self, 'EdaModbusSerialConf')
        configuration.Configurable.__init__(self)
        self.setModuleName(modname)
        dataSource.DataSource.__init__(self)
        self.setID(modname)

        self.portName = None
        self.mode = None
        self.unitId = None
        self.edaListener = None
        self.modbusSource = None
        self.offbyonefix = False
        
    def getServiceName(self):
        return self.getModuleName()

    def handleConfiguration(self, conf):
        self.portName = conf.getValue('SERIAL_PORT', '', self.getModuleName())
        if len(self.portName) <= 0:
            return (-1, 'Missing serial port name.')

        self.offbyonefix = conf.isTrue('OFFBYONEFIX', self.getModuleName())

        self.mode = conf.getValue('MODE', 'MASTER', self.getModuleName())
        if not self.mode in ['MASTER', 'SLAVE']:
            return (-1, 'Invalid Modbus RTU mode (allowed values: MASTER, SLAVE).')

        if self.mode == 'MASTER':
            try:
                self.unitId = int(conf.getValue('UNITID', '1', self.getModuleName()))
            except:
                return (-1, 'Cannot convert UNITID to integer.')

        return (1, '')
            
    def initConfiguration(self):        
        if self.mode == 'SLAVE':
            self.edaListener = enerventEda.EdaModbusRTUListener(self.portName)
            self.edaListener.setOffByOneFix(self.offbyonefix)
            if not self.edaListener.init():
                return (-1, 'Cannot initialize Enervent EDA modbus RTU listener.')
        elif self.mode == 'MASTER':            
            privateConf = configuration.Configuration()
            privateConf.setValue('TYPE', 'RTU', self.getModuleName())
            privateConf.setValue('SERIAL_PORT', self.portName, self.getModuleName())
            privateConf.setValue('BAUDRATE', '19200', self.getModuleName())
            privateConf.setValue('PARITY', 'NONE', self.getModuleName())
            privateConf.setValue('STOPBITS', '1', self.getModuleName())
            privateConf.setValue('DATABITS', '8', self.getModuleName())
            privateConf.setValue('RESPONSETIMEOUT', '3.0', self.getModuleName())

            coils = []        
            holdingregisters = []        
            for register in enerventEda.EDA_DEVICES['DEFAULT']:
                edaType = EDA_TYPE_MAP[register[3]]
                if self.offbyonefix:
                    address = register[1] + 1
                else:
                    address = register[1]
                if register[2] == enerventEda.TYPE_COIL:
                    coils.append("%s:%d:%d" % (register[0], self.unitId, address))
                elif register[2] == enerventEda.TYPE_HREG:
                    holdingregisters.append("%s:%d:%d:%s:%s" % (register[0], self.unitId, address, edaType[0], edaType[1]))
            
            privateConf.setValue('COILSTATE', coils, self.getModuleName())
            privateConf.setValue('HOLDINGREGISTER', holdingregisters, self.getModuleName())

            self.modbusSource = sourceModbus.ModbusConf(self.getModuleName())
    
            (stat, msg) = self.modbusSource.handleConfiguration(privateConf)
            if stat != 1:
                return (stat, msg)
            
            (stat, msg) = self.modbusSource.initConfiguration()
            if stat != 1:
                return (stat, msg)
        else:
            return (-1, 'Invalid Modbus RTU mode.')
        
        return (1, '')

    def releaseConfiguration(self):
        if self.edaListener:
            self.edaListener.close()
            self.edaListener = None

        if self.modbusSource != None:
            self.modbusSource.releaseConfiguration()
            self.modbusSource = None

        return (1, '')

    def runDataSourceQueryCommandImpl(self, cmds):
        if self.mode == 'MASTER':
            return self.modbusSource.runDataSourceQueryCommandImpl(cmds)
        elif self.mode == 'SLAVE':
            res = {}
            for cmd in cmds:
                if self.edaListener != None:
                    res[cmd] = self.edaListener.runQueryCommand(cmd)
                else:
                    res[cmd] = ''                                
            return res
        else:
            return {}
    
    @staticmethod
    def getModuleTypeName():
        return 'EDASERIAL'

    @staticmethod
    def getAllowedConfigurationKeys():
        return (['SERIAL_PORT', 'MODE', 'UNITID', 'OFFBYONEFIX'], [])
