#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            storeMQTT.py
#
# License:         Donationware, see attached LICENSE file for more 
#                  information
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         0.1a
#
# Date:            15.03.2020
#
# Description:     Data Store class to store logged data values into a
#                  ThingSpeak cloud (https://thingspeak.com).
#                  
# Requirements:    Python interpreter 2.7 or newer (www.python.org)
# 
# Version history: ** 15.03.2020 v0.1a (Olli Lammi) **
#                  First beta version for testers. 
#
###########################################################################

# Imports

import sys, os
import string, time
import json
import importlib

from modules.core import store
from modules.core import configuration

###########################################################################

# Constants

    
###########################################################################

# Classes

class MQTTStore(store.Store):
    def __init__(self, modname):
        store.Store.__init__(self, modname)

    def handleConfiguration(self, conf):
        if not self.initStoreFilters(conf):
            return (0, "Invalid store filter.")
        
        self.host = conf.getValue('HOST', '', self.getModuleName())
        try:
            self.port = int(conf.getValue('PORT', '1883', self.getModuleName()))
        except:
            return (0, 'Invalid MQTT broker port value.')

        self.client_id = conf.getValue('CLIENT_ID', '', self.getModuleName())
        self.username = conf.getValue('USERNAME', '', self.getModuleName())
        self.password = conf.getValue('PASSWORD', '', self.getModuleName())
        self.protocol = conf.getValue('PROTOCOL', '1', self.getModuleName())
        self.topic = conf.getValue('TOPIC', '', self.getModuleName())
        self.publish_scheme = conf.getValue('PUBLISH_SCHEME', '1', self.getModuleName())

        if not self.protocol in ['1', '2']:
            return (0, "Invalid MQTT protocol.")
        self.protocol = int(self.protocol)

        if not self.publish_scheme in ['1', '2', '3']:
            return (0, "Invalid publish scheme.")
        self.publish_scheme = int(self.publish_scheme)

        if len(self.host) <= 0:
            return (0, "Missing MQTT broker host.")

        if len(self.client_id) <= 0:
            return (0, "Missing MQTT client id.")

        if len(self.topic) <= 0:
            return (0, "Missing MQTT topic.")

        if self.topic[0] == '/':
            return (0, "MQTT topic should not start with /.")
        if self.topic[len(self.topic)-1] == '/':
            return (0, "MQTT topic should not end with /.")

        return (1, '')

    def initConfiguration(self):
        try:
            global mqttpub
            mqttpub = importlib.import_module('paho.mqtt.publish') 
            global mqttcli
            mqttcli = importlib.import_module('paho.mqtt.client') 
        except Exception, e:
            return (0, 'storeMQTT: Error loading Paho MQTT module paho.mqtt: %s' % e.__str__())

        return (1, '')

    def insertDataImpl(self, timeval, values):
        if self.protocol == 2:
            protocol = mqttcli.MQTTv311
        else:
            protocol = mqttcli.MQTTv31

        auth = None
        if len(self.username) > 0:
            auth = {'username': self.username, 'password': self.password}

        if self.publish_scheme == 1:
            msgs = []
            for val in values:
                self.Log(repr(val))
                topic = self.topic + '/' + val[0]
                payload = val[1]
                msgs.append({'topic': topic, 'payload': payload})
            try:
                self.Debug("Publishing MQTT data: %s" % repr(msgs))
                mqttpub.multiple(msgs, hostname=self.host, port=self.port, client_id=self.client_id, auth=auth, protocol=protocol, transport="tcp")
            except Exception, e:
                self.Log("ERROR: Error publishing data: %s" % e.__str__())
                return 0
        elif self.publish_scheme == 2:
            msgs = []
            for val in values:
                topic = self.topic + '/' + val[0]
                payload = {'timestamp': timeval, 'value': val[1]}
                payload = json.dumps(payload)
                msgs.append({'topic': topic, 'payload': payload})
            try:
                self.Debug("Publishing MQTT data: %s" % repr(msgs))
                mqttpub.multiple(msgs, hostname=self.host, port=self.port, client_id=self.client_id, auth=auth, protocol=protocol, transport="tcp")
            except Exception, e:
                self.Log("ERROR: Error publishing data: %s" % e.__str__())
                return 0
        elif self.publish_scheme == 3:
            topic = self.topic
            payload = {'timestamp': timeval}
            for val in values:
                payload[val[0]] = val[1]
            payload = json.dumps(payload)
            try:
                self.Debug("Publishing MQTT data topic: %s, payload: %s" % (topic, payload))
                mqttpub.single(topic, payload=payload, hostname=self.host, port=self.port, client_id=self.client_id, auth=auth, protocol=protocol, transport="tcp")
            except Exception, e:
                self.Log("ERROR: Error publishing data to topic: %s: %s" % (topic, e.__str__()))
                return 0
            
        return 1

    @staticmethod
    def getModuleTypeName():
        return 'MQTT'
    
    @staticmethod
    def getAllowedConfigurationKeys():
        return (['HOST', 'PORT', 'CLIENT_ID', 'USERNAME', 'PASSWORD', 'PROTOCOL', 'TOPIC', 'PUBLISH_SCHEME'], [])
