#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            storePostgreDb.py
#
# License:         Donationware, see attached LICENSE file for more 
#                  information
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         1.0b
#
# Date:            03.01.2017
#
# Description:     Data Store class to store logged data values into a
#                  PostgreSQL database.
#                  
# Requirements:    Python interpreter 2.4 or newer (www.python.org)
#                  (tested with 2.4.3)
# 
#                  Psycopg2-library. 
#                  (http://initd.org/psycopg/)
#
# Version history: ** 30.01.2016 v1.0a (Olli Lammi) **
#                  First version. 
#
#                  ** 03.01.2017 v1.0b (Olli Lammi) **
#                  Improved parameter error handling. 
#
###########################################################################

# Imports

import sys, os 
import string

from modules.core import store
from modules.core import configuration


# Psycopg2 module, not loaded until init of DBPostgreStore class instance
Psycopg2 = None


###########################################################################

# Constants

CREATE_TALO_DATA_1 = "CREATE TABLE IF NOT EXISTS talo_data (" + \
                     "id SERIAL PRIMARY KEY, " + \
                     "time TIMESTAMP WITH TIME ZONE NOT NULL, " + \
                     "position_id SMALLINT NOT NULL REFERENCES talo_positions(id), " + \
                     "value DECIMAL)"

CREATE_TALO_DATA_2 = "CREATE INDEX talo_data_time_idx ON talo_data (time)"

CHECK_FOR_TALO_DATA_INDEX = "SELECT indexname FROM pg_catalog.pg_indexes WHERE indexname = 'talo_data_time_idx' LIMIT 1"

CREATE_TALO_POSITIONS_1 = "CREATE TABLE IF NOT EXISTS talo_positions (" + \
                     "id SMALLSERIAL PRIMARY KEY, " + \
                     "position_name VARCHAR(32) NOT NULL UNIQUE)"


###########################################################################

# Classes

class DBPostgreStore(store.Store):
    def __init__(self, modname):
        store.Store.__init__(self, modname)
    
        self.POSITIONS = {}

    def handleConfiguration(self, conf):
        if not self.initStoreFilters(conf):
            return (0, "Invalid store filter.")
        
        self.DB_HOST = conf.getValue('HOST', '', self.getModuleName())
        try:
            self.DB_PORT = int(conf.getValue('PORT', '5432', self.getModuleName()))
        except:
            return (0, 'Invalid DB port value.')
        self.DB_USER = conf.getValue('USER', '', self.getModuleName())
        self.DB_PASSWD = conf.getValue('PASSWD', '', self.getModuleName())
        self.DB_NAME = conf.getValue('NAME', '', self.getModuleName())
        self.DB_TABLE = conf.getValue('TABLE', '', self.getModuleName())
        self.DB_TIMECOL = conf.getValue('TIMECOL', '', self.getModuleName())
                            
        if not len(self.DB_NAME) > 0:
            return (-1, 'Missing database name from database configuration.')
                
        self.DB_SCHEMA = conf.getValue('DB_SCHEMA_VERSION', '', self.getModuleName())
        if len(self.DB_SCHEMA) <= 0 or int(self.DB_SCHEMA) <= 0:
            self.DB_SCHEMA = 0
        elif len(self.DB_SCHEMA) > 0:
            self.DB_SCHEMA = int(self.DB_SCHEMA)
        if self.DB_SCHEMA == 1:
            self.DB_TABLE = 'talo_data'
            self.DB_TIMECOL = 'time'
        elif self.DB_SCHEMA == 0:
            if not len(self.DB_TABLE) > 0:
                return (-1, 'Missing database table name from database configuration.')
            if not len(self.DB_TIMECOL) > 0:
                return (-1, 'Missing database time column name from database configuration.')
        else:
            return (-1, 'Invalid DB schema version.')

        return (1, '')

    def initConfiguration(self):
        try:
            global Psycopg2
            Psycopg2 = __import__('psycopg2') 
        except Exception, e:
            self.Log("ERROR: Error loading database module psycopg2")
            return (0, 'DBPostgreStore: Error loading databse module psycopg2')
        
        status = 0
        sqlstmt = ""
        try:
            db = Psycopg2.connect(host=self.DB_HOST, port=self.DB_PORT, user=self.DB_USER, \
                                 password=self.DB_PASSWD, database=self.DB_NAME)
            cur = db.cursor()
            if self.DB_SCHEMA == 1:
                sqlstmt = CREATE_TALO_POSITIONS_1
                cur.execute(sqlstmt)
                sqlstmt = CREATE_TALO_DATA_1
                cur.execute(sqlstmt)
                sqlstmt = CHECK_FOR_TALO_DATA_INDEX
                cur.execute(sqlstmt)
                res = cur.fetchall()
                if len(res) <= 0:                    
                    sqlstmt = CREATE_TALO_DATA_2
                    cur.execute(sqlstmt)
            sqlstmt = "SELECT " + self.DB_TIMECOL + " FROM " + self.DB_TABLE + " LIMIT 1"
            cur.execute(sqlstmt)
            cur.close()
            db.commit()
            db.close()

            status = 1
        except Exception, e:
            self.Log("ERROR: Error in database operation, SQL: %s (%s)" % (sqlstmt, e.__str__()))
            status = 0
        if status:
            self.initPositions()
            return (1, '')
        else:
            return (0, 'DBPostgreStore: Errors with database access.')

    def initPositions(self):
        if self.DB_SCHEMA == 1:
            self.POSITIONS = {}
            sqlstmt = ""
            try:
                db = Psycopg2.connect(host=self.DB_HOST, port=self.DB_PORT, user=self.DB_USER, \
                                     password=self.DB_PASSWD, database=self.DB_NAME)
                cur = db.cursor()
                sqlstmt = "SELECT id, position_name FROM talo_positions"
                cur.execute(sqlstmt)
                res = cur.fetchall()
                cur.close()
                db.close()
                for row in res:
                    self.POSITIONS[row[1]] = row[0]
            except:
                self.Log("ERROR: Error in database operation, SQL: " + sqlstmt)
 
    def insertDataImpl(self, timeval, values):
        if self.DB_SCHEMA == 1:
            if len(self.POSITIONS.keys()) <= 0:
                self.initConfiguration()
            
            status = 1
            for val in values:
                if not self.POSITIONS.has_key(val[0]):
                    status = 0
                    break
            if status == 0:
                self.initPositions()
                toadd = []
                for val in values:
                    if not self.POSITIONS.has_key(val[0]):
                        toadd.append(val[0])
                                
                try:
                    db = Psycopg2.connect(host=self.DB_HOST, port=self.DB_PORT, user=self.DB_USER, \
                                          password=self.DB_PASSWD, database=self.DB_NAME)
                    cur = db.cursor()
                    for col in toadd:
                        sqlstmt = "INSERT INTO talo_positions (position_name) VALUES ('"
                        sqlstmt = sqlstmt + col
                        sqlstmt = sqlstmt + "')"
                        cur.execute(sqlstmt)
                    cur.close()
                    db.commit()
                    db.close()
                except:
                    self.Log("ERROR: Error in database operation, SQL: " + sqlstmt)
                    self.POSITIONS = {}
                    return 0
                self.initPositions()

        sqlstmt = ""
        try:
            db = Psycopg2.connect(host=self.DB_HOST, port=self.DB_PORT, user=self.DB_USER, \
                                  password=self.DB_PASSWD, database=self.DB_NAME)

            if self.DB_SCHEMA == 1:
                timestr = "TO_TIMESTAMP(%d)" % (timeval, )

                for val in values:
                    cur = db.cursor()
    
                    posid = self.POSITIONS[val[0]]
                            
                    sqlstmt = "INSERT INTO " + self.DB_TABLE + " "
                    sqlstmt = sqlstmt + "(" + self.DB_TIMECOL + ", position_id, value) VALUES ("
                    sqlstmt = sqlstmt + "%s, %d, " % (timestr, posid)
                    sqlstmt = sqlstmt + checkDBValue(val[1]) + ")"
        
                    self.Debug("Inserting data to database using SQL: " + sqlstmt)
                    cur.execute(sqlstmt)
                    cur.close()
            else:
                timestr = "TO_TIMESTAMP(%d)" % (timeval, )
                
                cur = db.cursor()
                sqlstmt = "SELECT " + self.DB_TIMECOL + " FROM " + self.DB_TABLE
                sqlstmt = sqlstmt + " WHERE " + self.DB_TIMECOL + " = " + timestr + " LIMIT 1"
                cur.execute(sqlstmt)
                res = cur.fetchall()
                cur.close()
                
                if len(res) <= 0:
                    cur = db.cursor()
                    sqlstmt = "INSERT INTO " + self.DB_TABLE + " "
                    sqlstmt = sqlstmt + "(" + self.DB_TIMECOL 
                    sqlstmt = sqlstmt + ") VALUES (" + timestr 
                    sqlstmt = sqlstmt + ")"
                    cur.execute(sqlstmt)
                    cur.close()                

                cur = db.cursor()    
                valuestrs = []
                for item in values:
                    valuestrs.append(item[0] + ' = ' + checkDBValue(item[1]))
                        
                sqlstmt = "UPDATE " + self.DB_TABLE + " SET "
                sqlstmt = sqlstmt + string.join(valuestrs, ", ")
                sqlstmt = sqlstmt + " WHERE " + self.DB_TIMECOL + " = " + timestr
    
                self.Debug("Setting data to database using SQL: " + sqlstmt)
                cur.execute(sqlstmt)
                cur.close()
            db.commit()
            db.close()
        except:
            self.Log("ERROR: Error in database operation, SQL: " + sqlstmt)
            self.POSITIONS = {}
            return 0
        return 1

    @staticmethod
    def getModuleTypeName():
        return 'POSTGREDB'

    @staticmethod
    def getAllowedConfigurationKeys():
        return (['HOST', 'PORT', 'USER', 'PASSWD', 'NAME', 'TABLE', 'TIMECOL', 'DB_SCHEMA_VERSION'], [])


###########################################################################

# Functions

def checkDBValue(val):
    if (val and len(val) > 0):
        return val
    return "NULL"

